@@ -60,3 +60,20 @@ class UnsatisfiedSchemaError(Exception):
 
 class TaskElementExecutionError(Exception):
     'For when the execution of an task element fails.'
+
+
+class ConfigurationError(Exception):
+    'For malformed configuration files.'
+
+
+class SoftwareInstanceError(Exception):
+    'For malformed SoftwareInstance definitions.'
+    pass
+
+
+class MissingSoftwareSourcesError(Exception):
+    'For when a software instance requires source variables, but none are forthcoming.'
+
+
+class UnexpectedSourceMapReturnError(Exception):
+    'For when a source map function does not return the expected dict.'
