@@ -14,8 +14,9 @@ import copy
 from warnings import warn
 
 import numpy as np
+from hpcflow.scheduler import SunGridEngine
 
-from matflow import SOFTWARE, TASK_SCHEMAS, SCHEMA_IS_VALID
+from matflow.config import Config
 from matflow.errors import (
     IncompatibleWorkflow,
     MissingSoftware,
@@ -28,10 +29,13 @@ from matflow.errors import (
     UnsatisfiedGroupParameter,
     MissingSchemaError,
     UnsatisfiedSchemaError,
+    MissingSoftwareSourcesError,
 )
 from matflow.utils import (tile, repeat, arange, extend_index_list, flatten_list,
-                           to_sub_list)
+                           to_sub_list, get_specifier_dict)
 from matflow.models.task import Task, TaskSchema
+from matflow.models.element import Element
+from matflow.models.software import SoftwareInstance
 
 
 def normalise_local_inputs(base=None, sequences=None):
@@ -193,22 +197,86 @@ def get_local_inputs(schema_inputs, base=None, num_repeats=1, sequences=None,
     return local_ins
 
 
-def get_software_instance(software_name, num_cores, all_software):
+def get_software_instance(software, run_options, all_software, type_label=''):
     """Find a software instance in the software.yml file that matches the software
-    requirements of a given task."""
+    requirements of a given task.
 
-    for soft_inst in all_software:
+    Parameters
+    ----------
+    software : dict
+        Dict with the following keys:
+            name : str
+                Name of the software whose SoftwareInstance is to be returned.
+            label : str
+                Additional specifier used to distinguish, e.g., a software version.
+            options : list of str
+                Additional specifiers used to state additional requirements.
+    run_options : dict
+        Dict with keys:
+            num_cores : int
+                Number of cores specified in the task.
+            **scheduler_options
+                Any other options to be passed to the scheduler.
+    all_software : dict of list of SoftwareInstance
+        Dict whose keys are software names and whose values are lists of SoftwareInstance
+        objects.
 
-        if soft_inst['name'] != software_name:
+    Returns
+    -------
+    SoftwareInstance
+        The first matching software instance.
+
+    Raises
+    ------
+    MissingSoftware
+        If no matching software instance can be found.
+
+    """
+    match = None
+    for name, instances in all_software.items():
+
+        if name != SoftwareInstance.get_software_safe(software['name']):
             continue
 
-        core_range = soft_inst['num_cores']
-        all_num_cores = list(range(*core_range)) + [core_range[1]]
-        if num_cores in all_num_cores:
-            return soft_inst
+        for inst in instances:
 
-    raise MissingSoftware(f'Could not find suitable software "{software_name}", with'
-                          f' `num_cores={num_cores}`.')
+            if run_options['num_cores'] not in inst.cores_range:
+                continue
+            if inst.label != software['label']:
+                continue
+            if (set(software['options']) - set(inst.options)):
+                continue
+
+            # Check no conflicting scheduler options
+            keep_looking = False
+            for k, v in inst.required_scheduler_options.items():
+                if k in run_options and v != run_options[k]:
+                    keep_looking = True
+                    break
+
+            if keep_looking:
+                continue
+            else:
+                match = inst
+                break
+
+        if match:
+            break
+
+    if match:
+        if run_options['num_cores'] > 1:
+            all_run_opts = {**match.required_scheduler_options, **run_options}
+            # (SGE specific):
+            if 'pe' not in all_run_opts:
+                msg = ('Parallel environment (`pe`) key must be specified in '
+                       f'`run_options{type_label}`, since `num_cores > 1`.')
+                raise TaskError(msg)
+    else:
+        msg = (f'Could not find suitable software "{software["name"]}", with '
+               f'`num_cores={run_options["num_cores"]}` and `label={software["label"]}`.')
+        raise MissingSoftware(msg)
+
+    return match
 
 
 def get_dependency_idx(task_info_lst):
@@ -239,6 +307,7 @@ def get_dependency_idx(task_info_lst):
               - The upstream and downstream task share the same context, and,
                 for any downstream task input parameter, the parameter context
                 is `None`.
+              - The upstream task context is default (i.e. '').
 
     """
 
@@ -270,7 +339,7 @@ def get_dependency_idx(task_info_lst):
                     param_context == upstream_context or (
                         (upstream_context == downstream_context) and
                         (param_context is None)
-                    )
+                    ) or (upstream_context == '')
                 ):
                     output_idx.append(task_idx_k)
 
@@ -293,8 +362,42 @@ def get_dependency_idx(task_info_lst):
     return dependency_idx
 
 
+def validate_run_options(run_opts, type_label=''):
+
+    # SGE specific:
+    ALLOWED = SunGridEngine.ALLOWED_USER_OPTS + ['num_cores']
+    if 'preparation' in type_label or 'processing' in type_label:
+        ALLOWED += ['job_array']
+
+    bad_keys = set(run_opts.keys()) - set(ALLOWED)
+    if bad_keys:
+        bad_keys_fmt = ', '.join([f'{i!r}' for i in bad_keys])
+        raise TaskError(f'Run options not known: {bad_keys_fmt}.')
+
+    run_opts = copy.deepcopy(run_opts)
+
+    if 'preparation' in type_label:
+        run_opts = {**Config.get('default_preparation_run_options'), **run_opts}
+    elif 'processing' in type_label:
+        run_opts = {**Config.get('default_processing_run_options'), **run_opts}
+
+    if 'num_cores' not in run_opts:
+        run_opts.update({'num_cores': 1})
+
+    if run_opts['num_cores'] <= 0:
+        msg = (f'Specify `num_cores` (in `run_options{type_label}`) as an integer '
+               f'greater than 0.')
+        raise TaskError(msg)
+    elif 'pe' in run_opts:
+        msg = (f'No need to specify parallel environment (`pe`) in '
+               f'`run_options{type_label}`, since `num_cores=1`.')
+        raise TaskError(msg)
+
+    return run_opts
+
+
 def validate_task_dict(task, is_from_file, all_software, all_task_schemas,
-                       check_integrity=True):
+                       all_sources_maps, check_integrity=True):
     """Validate a task dict.
 
     Parameters
@@ -306,9 +409,13 @@ def validate_task_dict(task, is_from_file, all_software, all_task_schemas,
     is_from_file : bool
         Has this task dict been loaded from a workflow file or is it associated
         with a brand new workflow?
-    all_software
+    all_software : dict of list of SoftwareInstance
+        Dict whose keys are software names and whose values are lists of SoftwareInstance
+        objects.
     all_task_schemas : dict of (tuple : TaskSchema)
         All available TaskSchema objects, keyed by a (name, method, software) tuple.
+    all_sources_maps : dict of (tuple : dict)
+        All available sources maps.
     check_integrity : bool, optional
         Applicable if `is_from_file` is True. If True, re-generate `local_inputs`
         and compare them to those loaded from the file. If the equality test
@@ -317,9 +424,12 @@ def validate_task_dict(task, is_from_file, all_software, all_task_schemas,
 
     Returns
     -------
-    task : dict
+    task_list : list of dict
         Copy of original `task` dict, with default keys added. If `is_from_file=False`,
-        `local_inputs` and `schema` are added.
+        `local_inputs` and `schema` are added. Usually this will just be a list of length
+        one. However, there are some scenarios where multiple task dicts will be returned.
+        For example, if multiple `contexts` are specified, the task will be repeated,
+        once for each `context`.
 
     """
 
@@ -331,17 +441,19 @@ def validate_task_dict(task, is_from_file, all_software, all_task_schemas,
             'id',
             'name',
             'method',
+            'elements',
             'software_instance',
+            'prepare_software_instance',
+            'process_software_instance',
             'task_idx',
             'run_options',
+            'prepare_run_options',
+            'process_run_options',
             'status',
             'stats',
             'context',
             'local_inputs',
-            'inputs',
-            'outputs',
             'schema',
-            'files',
             'resource_usage',
             'base',
             'sequences',
@@ -360,6 +472,7 @@ def validate_task_dict(task, is_from_file, all_software, all_task_schemas,
             'run_options',
             'stats',
             'context',
+            'contexts',
             'base',
             'sequences',
             'repeats',
@@ -372,7 +485,6 @@ def validate_task_dict(task, is_from_file, all_software, all_task_schemas,
         def_keys = {
             'run_options': {},
             'stats': True,
-            'context': '',
             'base': None,
             'sequences': None,
             'repeats': 1,
@@ -397,24 +509,13 @@ def validate_task_dict(task, is_from_file, all_software, all_task_schemas,
         raise TaskError(msg)
 
     task = {**def_keys, **copy.deepcopy(task)}
-    if 'num_cores' not in task['run_options']:
-        task['run_options'].update({'num_cores': 1})
 
-    # (SGE specific):
-    if task['run_options']['num_cores'] > 1:
-        if 'pe' not in task['run_options']:
-            msg = ('Parallel environment (`pe`) key must be specified in `run_options`, '
-                   'since `num_cores > 1`.')
-            raise TaskError(msg)
-
-    elif task['run_options']['num_cores'] <= 0:
-        msg = 'Specify `num_cores` (in `run_options`) as an integer greater than 0.'
-        raise TaskError(msg)
-
-    elif 'pe' in task['run_options']:
-        msg = ('No need to specify parallel environment (`pe`) in `run_options`, since '
-               '`num_cores=1`.')
-        raise TaskError(msg)
+    all_run_opts = task.pop('run_options')
+    prep_run_opts = all_run_opts.pop('preparation', {})
+    proc_run_opts = all_run_opts.pop('processing', {})
+    task['prepare_run_options'] = validate_run_options(prep_run_opts, '.preparation')
+    task['process_run_options'] = validate_run_options(proc_run_opts, '.processing')
+    task['run_options'] = validate_run_options(all_run_opts)
 
     # Make TaskSchema:
     if is_from_file:
@@ -422,25 +523,71 @@ def validate_task_dict(task, is_from_file, all_software, all_task_schemas,
         schema = TaskSchema(**task['schema'])
 
     else:
+
+        # Normalise for multiple `contexts`:
+        if 'context' in task and 'contexts' in task:
+            msg = ('Specify exactly one of `context` and `contexts` (these keys are '
+                   'equivalent).')
+            raise TaskError(msg)
+
+        elif ('context' not in task) and ('contexts' not in task):
+            task['contexts'] = ''
+
+        elif 'context' in task:
+            task['contexts'] = task.pop('context')
+
+        if not isinstance(task['contexts'], list):
+            task['contexts'] = [task['contexts']]
+
         # Find the software instance:
-        soft_inst = get_software_instance(
+        software = get_specifier_dict(
             task.pop('software'),
-            task['run_options']['num_cores'],
+            name_key='name',
+            base_key='label',
+            list_specifiers=['options'],
+            defaults={'label': None, 'options': []},
+        )
+
+        soft_inst = get_software_instance(
+            software,
+            task['run_options'],
             all_software,
         )
+        prepare_soft_inst = get_software_instance(
+            software,
+            task['prepare_run_options'],
+            all_software,
+            type_label='.prepare',
+        )
+        process_soft_inst = get_software_instance(
+            software,
+            task['process_run_options'],
+            all_software,
+            type_label='.process',
+        )
+
+        # print(f'prepare_soft_inst:\n{prepare_soft_inst}')
+        # print(f'process_soft_inst:\n{process_soft_inst}')
+
+        schema_key = (task['name'], task['method'], soft_inst.software)
+        # Check any sources required by the main software instance are defined in the
+        # sources map:
+        soft_inst.validate_source_maps(*schema_key, all_sources_maps)
+
         task['software_instance'] = soft_inst
+        task['prepare_software_instance'] = prepare_soft_inst
+        task['process_software_instance'] = process_soft_inst
 
         # Find the schema:
-        schema_key = (task['name'], task['method'], soft_inst['name'])
-        if not SCHEMA_IS_VALID[schema_key]:
-            msg = (f'No matching extension function found for the schema with '
-                   f'implementation: {soft_inst["name"]}.')
-            raise UnsatisfiedSchemaError(msg)
         schema = all_task_schemas.get(schema_key)
         if not schema:
             msg = (f'No matching task schema found for task name "{task["name"]}" with '
-                   f'method "{task["method"]}" and software "{soft_inst["name"]}".')
+                   f'method "{task["method"]}" and software "{soft_inst.software}".')
             raise MissingSchemaError(msg)
+        if not Config.get('schema_validity')[schema_key]:
+            msg = (f'No matching extension function found for the schema with '
+                   f'implementation: {soft_inst.software}.')
+            raise UnsatisfiedSchemaError(msg)
 
     local_ins = get_local_inputs(
         schema.input_names,
@@ -453,7 +600,13 @@ def validate_task_dict(task, is_from_file, all_software, all_task_schemas,
     )
 
     if is_from_file and check_integrity:
-        if local_ins != task['local_inputs']:
+
+        # Don't compare the vals_data_idx:
+        loaded_local_inputs = copy.deepcopy(task['local_inputs'])
+        for vals_dict in loaded_local_inputs['inputs'].values():
+            del vals_dict['vals_data_idx']
+
+        if local_ins != loaded_local_inputs:
             msg = (
                 f'Regenerated local inputs (task: "{task["name"]}") '
                 f'are not equivalent to those loaded from the '
@@ -463,10 +616,23 @@ def validate_task_dict(task, is_from_file, all_software, all_task_schemas,
             )
             raise WorkflowPersistenceError(msg)
 
+        local_ins = task['local_inputs']
+
     task['local_inputs'] = local_ins
     task['schema'] = schema
 
-    return task
+    if is_from_file:
+        task_list = [task]
+
+    else:
+        task_list = []
+        for context in task['contexts']:
+            task_copy = copy.deepcopy(task)
+            task_copy['context'] = context
+            del task_copy['contexts']
+            task_list.append(task_copy)
+
+    return task_list
 
 
 def check_consistent_inputs(task_lst, dep_idx):
@@ -893,7 +1059,7 @@ def get_element_idx(task_lst, dep_idx):
     return element_idx
 
 
-def init_tasks(task_lst, is_from_file, check_integrity=True):
+def init_tasks(workflow, task_lst, is_from_file, check_integrity=True):
     """Construct and validate Task objects and the element indices
     from which to populate task inputs.
 
@@ -924,8 +1090,16 @@ def init_tasks(task_lst, is_from_file, check_integrity=True):
 
     # Validate and add `schema` and `local_inputs` to each task:
     task_lst = [
-        validate_task_dict(i, is_from_file, SOFTWARE, TASK_SCHEMAS, check_integrity)
+        j
         for i in task_lst
+        for j in validate_task_dict(
+            i,
+            is_from_file,
+            Config.get('software'),
+            Config.get('task_schemas'),
+            Config.get('sources_maps'),
+            check_integrity
+        )
     ]
 
     # Get dependencies, sort and add `task_idx` to each task:
@@ -935,10 +1109,19 @@ def init_tasks(task_lst, is_from_file, check_integrity=True):
     element_idx = get_element_idx(task_lst, dep_idx)
 
     task_objs = []
-    for i in task_lst:
+    for task_idx, task_dict in enumerate(task_lst):
+
+        if is_from_file:
+            task_id = task_dict.pop('id')
+            elements = task_dict.pop('elements')
+        else:
+            task_id = None
+            num_elements = element_idx[task_idx]['num_elements']
+            elements = [{'element_idx': elem_idx} for elem_idx in range(num_elements)]
+
+        task = Task(workflow=workflow, **task_dict)
+        task.init_elements(elements)
 
-        task_id = i.pop('id') if is_from_file else None
-        task = Task(**i)
         if is_from_file:
             task.id = task_id
         else:
