@@ -4,6 +4,7 @@ Module containing the Task and TaskSchema classes.
 
 """
 
+import enum
 import re
 import secrets
 from pprint import pprint
@@ -12,23 +13,18 @@ import numpy as np
 
 from matflow.models import CommandGroup
 from matflow.errors import TaskSchemaError, TaskParameterError
+from matflow.hicklable import to_hicklable
+from matflow.utils import dump_to_yaml_string, get_specifier_dict
+from matflow.models.software import SoftwareInstance
+from matflow.models.element import Element
 
 
 class TaskSchema(object):
-    """Class to represent the schema of a particular method/implementation of a task.
-
-            'name': self.name,
-            'method': self.method,
-            'implementation': self.software,
-            'inputs': inputs,
-            'outputs': outputs,
-            'input_map': in_map,
-            'output_map': out_map,
-            'command_group': command_group,
-    """
+    """Class to represent the schema of a particular method/implementation of a task."""
 
     def __init__(self, name, outputs, method=None, implementation=None, inputs=None,
-                 input_map=None, output_map=None, command_group=None):
+                 input_map=None, output_map=None, command_group=None,
+                 archive_excludes=None):
         'Instantiate a TaskSchema.'
 
         self.name = name
@@ -38,27 +34,113 @@ class TaskSchema(object):
         self.inputs = inputs or []
         self.input_map = input_map or []
         self.output_map = output_map or []
+        self.archive_excludes = archive_excludes
 
         self.command_group = CommandGroup(**command_group) if command_group else None
 
         self._validate_inputs_outputs()
 
+    def __repr__(self):
+        return (
+            f'{self.__class__.__name__}('
+            f'name={self.name!r}, '
+            f'method={self.method!r}, '
+            f'implementation={self.implementation!r}, '
+            f'inputs={self.inputs_condensed!r}, '
+            f'outputs={self.outputs!r}, '
+            f'command_group={self.command_group!r}, '
+            f'archive_excludes={self.archive_excludes!r}'
+            f')'
+        )
+
+    def __str__(self):
+        self_dict = self.as_dict()
+        self_dict['inputs'] = self.inputs_condensed
+        return dump_to_yaml_string(self_dict)
+
+    def as_dict(self):
+        return to_hicklable(self)
+
     @classmethod
     def load_from_hierarchy(cls, schema_lst):
 
+        REQ = ['name', 'methods']
+        ALLOWED = REQ + ['inputs', 'outputs', 'notes']
+
+        REQ_METHOD = ['name', 'implementations']
+        ALLOWED_METHOD = REQ_METHOD + ['inputs', 'outputs', 'notes']
+
+        REQ_IMPL = ['name']
+        ALLOWED_IMPL = REQ_IMPL + [
+            'inputs',
+            'outputs',
+            'input_map',
+            'output_map',
+            'commands',
+            'notes',
+            'archive_excludes',
+        ]
+
         all_schema_dicts = {}
         for schema in schema_lst:
 
-            name = schema['name']
+            name = schema.get('name', 'MISSING NAME')
+
+            bad_keys = set(schema.keys()) - set(ALLOWED)
+            miss_keys = set(REQ) - set(schema.keys())
+
+            if bad_keys:
+                bad_keys_fmt = ', '.join(['"{}"'.format(i) for i in bad_keys])
+                msg = (f'Unknown task schema keys for task schema "{name}": '
+                       f'{bad_keys_fmt}.')
+                raise TaskSchemaError(msg)
+
+            if miss_keys:
+                miss_keys_fmt = ', '.join(['"{}"'.format(i) for i in miss_keys])
+                msg = (f'Missing task schema keys for task schema "{name}": '
+                       f'{miss_keys_fmt}.')
+                raise TaskSchemaError(msg)
 
             for method in schema['methods']:
 
+                bad_keys = set(method.keys()) - set(ALLOWED_METHOD)
+                miss_keys = set(REQ_METHOD) - set(method.keys())
+
+                if bad_keys:
+                    bad_keys_fmt = ', '.join(['"{}"'.format(i) for i in bad_keys])
+                    msg = (f'Unknown task schema method keys for task schema '
+                           f'"{name}": {bad_keys_fmt}.')
+                    raise TaskSchemaError(msg)
+
+                if miss_keys:
+                    miss_keys_fmt = ', '.join(['"{}"'.format(i) for i in miss_keys])
+                    msg = (f'Missing task schema method keys for task schema '
+                           f'"{name}": {miss_keys_fmt}.')
+                    raise TaskSchemaError(msg)
+
                 for imp in method['implementations']:
 
-                    key = (name, method['name'], imp['name'])
+                    bad_keys = set(imp.keys()) - set(ALLOWED_IMPL)
+                    miss_keys = set(REQ_IMPL) - set(imp.keys())
+
+                    if bad_keys:
+                        bad_keys_fmt = ', '.join(['"{}"'.format(i) for i in bad_keys])
+                        msg = (f'Unknown task schema implementation keys for task schema '
+                               f'"{name}" and method "{method["name"]}": {bad_keys_fmt}.')
+                        raise TaskSchemaError(msg)
+
+                    if miss_keys:
+                        miss_keys_fmt = ', '.join(['"{}"'.format(i) for i in miss_keys])
+                        msg = (f'Missing task schema method keys for task schema '
+                               f'"{name}" and method "{method["name"]}": '
+                               f'{miss_keys_fmt}.')
+                        raise TaskSchemaError(msg)
+
+                    software = SoftwareInstance.get_software_safe(imp['name'])
+                    key = (name, method['name'], software)
                     if key in all_schema_dicts:
                         msg = (f'Schema with name "{name}", method "{method["name"]}" '
-                               f'and implementation "{imp["name"]}" is multiply defined.')
+                               f'and implementation "{software}" is multiply defined.')
                         raise ValueError(msg)
 
                     input_map = imp.get('input_map', [])
@@ -78,12 +160,13 @@ class TaskSchema(object):
                         key: {
                             'name': name,
                             'method': method['name'],
-                            'implementation': imp['name'],
+                            'implementation': software,
                             'inputs': all_inputs,
                             'outputs': all_outputs,
                             'input_map': input_map,
                             'output_map': output_map,
                             'command_group': command_group,
+                            'archive_excludes': imp.get('archive_excludes'),
                         }
                     })
 
@@ -103,10 +186,28 @@ class TaskSchema(object):
     def input_contexts(self):
         return list(set([i.get('context', None) for i in self.inputs]))
 
+    @property
+    def inputs_condensed(self):
+        'Get inputs list in their string format.'
+        out = []
+        for i in self.inputs:
+            extra = ''
+            i_fmt = f'{i["name"]}'
+            if i['alias'] != i['name']:
+                extra += f'alias={i["alias"]}'
+            if i['context']:
+                extra += f'context={i["context"]}'
+            if i['group'] != 'default':
+                extra += f'group={i["group"]}'
+            if extra:
+                i_fmt += f'[{extra}]'
+            out.append(i_fmt)
+        return out
+
     def _validate_inputs_outputs(self):
         'Basic checks on inputs and outputs.'
 
-        allowed_inp_specifiers = ['group', 'context', 'alias']
+        allowed_inp_specifiers = ['group', 'context', 'alias', 'file']
         req_inp_keys = ['name']
         allowed_inp_keys = req_inp_keys + allowed_inp_specifiers
         allowed_inp_keys_fmt = ', '.join(['"{}"'.format(i) for i in allowed_inp_keys])
@@ -117,41 +218,14 @@ class TaskSchema(object):
         # Normalise schema inputs:
         for inp_idx, inp in enumerate(self.inputs):
 
-            if isinstance(inp, str):
-
-                # Parse additional input specifiers:
-                match = re.search(r'([\w-]+)(\[(.*?)\])*', inp)
-                inp_name = match.group(1)
-                inp = {'name': inp_name}
-
-                specifiers_str = match.group(3)
-                if specifiers_str:
-                    specs = specifiers_str.split(',')
-                    for s in specs:
-                        s_key, s_val = s.split('=')
-                        inp.update({s_key.strip(): s_val.strip()})
-
-                    if 'context' in inp and inp['context'] and 'alias' not in inp:
-                        msg = ('Task schema inputs for which a `context` is specified '
-                               'must also be given an `alias`.')
-                        raise TaskSchemaError(err + msg)
-
-            elif not isinstance(inp, dict):
-                raise TypeError(err + 'Task schema input must be a str or a dict.')
+            inp_defs = {'context': None, 'group': 'default', 'file': False}
+            inp = get_specifier_dict(inp, name_key='name', defaults=inp_defs)
 
             for r in req_inp_keys:
                 if r not in inp:
                     msg = f'Task schema input must include key {r}.'
                     raise TaskSchemaError(err + msg)
 
-            if 'context' not in inp:
-                # Add default parameter context:
-                inp['context'] = None
-
-            if 'group' not in inp:
-                # Add default group:
-                inp['group'] = 'default'
-
             if 'alias' not in inp:
                 # Add default alias:
                 inp['alias'] = inp['name']
@@ -173,11 +247,21 @@ class TaskSchema(object):
 
         # Check correct keys in supplied input/output maps:
         for in_map in self.input_map:
-            if sorted(in_map.keys()) != sorted(['inputs', 'file']):
-                bad_keys_fmt = ', '.join(['"{}"'.format(i) for i in in_map.keys()])
-                msg = (f'Input maps must map a list of `inputs` into a `file` but found '
-                       f'input map with keys {bad_keys_fmt}.')
-                raise TaskSchemaError(err + msg)
+
+            req_keys = ['inputs', 'file']
+            allowed_keys = set(req_keys + ['save'])
+            miss_keys = list(set(req_keys) - set(in_map.keys()))
+            bad_keys = list(set(in_map.keys()) - allowed_keys)
+
+            msg = (f'Input maps must map a list of `inputs` into a `file` (with an '
+                   f'optional `save` key).')
+            if miss_keys:
+                miss_keys_fmt = ', '.join(['"{}"'.format(i) for i in miss_keys])
+                raise TaskSchemaError(err + msg + f'Missing keys are: {miss_keys_fmt}.')
+            if bad_keys:
+                bad_keys_fmt = ', '.join(['"{}"'.format(i) for i in bad_keys])
+                raise TaskSchemaError(err + msg + f'Unknown keys are: {bad_keys_fmt}.')
+
             if not isinstance(in_map['inputs'], list):
                 msg = 'Input map `inputs` must be a list.'
                 raise TaskSchemaError(err + msg)
@@ -277,14 +361,6 @@ class TaskSchema(object):
     def is_func(self):
         return not self.command_group.commands
 
-    def __repr__(self):
-        out = (
-            f'{self.__class__.__name__}('
-            f'name={self.name!r}, method={self.method!r}, '
-            f'inputs={self.inputs!r}, outputs={self.outputs!r})'
-        )
-        return out
-
     def get_input_by_name(self, input_name):
         for i in self.inputs:
             if i['name'] == input_name:
@@ -298,6 +374,13 @@ class TaskSchema(object):
         raise ValueError(f'No input alias "{input_alias}" in schema.')
 
 
+class TaskStatus(enum.Enum):
+
+    pending = 1
+    running = 2
+    complete = 3
+
+
 class Task(object):
     """
 
@@ -309,24 +392,23 @@ class Task(object):
 
     """
 
-    INIT_STATUS = 'pending'
-
     __slots__ = [
         '_id',
         '_name',
         '_method',
         '_software_instance',
+        '_prepare_software_instance',
+        '_process_software_instance',
         '_task_idx',
         '_run_options',
+        '_prepare_run_options',
+        '_process_run_options',
         '_status',
         '_stats',
         '_context',
         '_local_inputs',
-        '_inputs',
-        '_outputs',
         '_output_map_options',
         '_schema',
-        '_files',
         '_resource_usage',
         '_base',
         '_sequences',
@@ -334,30 +416,36 @@ class Task(object):
         '_groups',
         '_nest',
         '_merge_priority',
+        '_workflow',
+        '_elements',
     ]
 
-    def __init__(self, name, method, software_instance, task_idx, run_options=None,
-                 status=None, stats=True, context='', local_inputs=None, inputs=None,
-                 outputs=None, schema=None, files=None, resource_usage=None,
-                 base=None, sequences=None, repeats=None, groups=None, nest=None,
-                 merge_priority=None, output_map_options=None):
+    def __init__(self, workflow, name, method, software_instance,
+                 prepare_software_instance, process_software_instance, task_idx,
+                 run_options=None, prepare_run_options=None, process_run_options=None,
+                 status=None, stats=True, context='', local_inputs=None, schema=None,
+                 resource_usage=None, base=None, sequences=None, repeats=None,
+                 groups=None, nest=None, merge_priority=None, output_map_options=None):
 
-        self._id = None  # Generated once by generate_id()
+        self._id = None         # Generated once by generate_id()
+        self._elements = None   # Assigned in init_elements()
 
+        self._workflow = workflow
         self._name = name
         self._method = method
         self._software_instance = software_instance
+        self._prepare_software_instance = prepare_software_instance
+        self._process_software_instance = process_software_instance
         self._task_idx = task_idx
-        self._run_options = run_options
-        self._status = status or Task.INIT_STATUS  # | 'paused' | 'complete'
+        self._run_options = run_options or {}
+        self._prepare_run_options = prepare_run_options or {}
+        self._process_run_options = process_run_options or {}
+        self._status = status or TaskStatus.pending
         self._stats = stats
         self._context = context
         self._local_inputs = local_inputs
-        self._inputs = inputs
-        self._outputs = outputs
         self._output_map_options = output_map_options
         self._schema = schema
-        self._files = files
         self._resource_usage = resource_usage
 
         # Saved for completeness, and to allow regeneration of `local_inputs`:
@@ -369,25 +457,55 @@ class Task(object):
         self._merge_priority = merge_priority
 
     def __repr__(self):
+
         out = (
             f'{self.__class__.__name__}('
             f'name={self.name!r}, '
-            f'status={self.status!r}, '
             f'method={self.method!r}, '
-            f'software_instance={self.software_instance!r}, '
-            f'run_options={self.run_options!r}, '
-            f'context={self.context!r}, '
-            f'schema={self.schema!r}'
+            f'software={self.software!r}'
             f')'
         )
         return out
 
+    def __str__(self):
+        return (
+            f'{"ID:":10}{self.id!s}\n'
+            f'{"Index:":10}{self.task_idx!s}\n'
+            f'{"Status:":10}{self.status!s}\n'
+            f'{"Name:":10}{self.name!s}\n'
+            f'{"Method:":10}{self.method!s}\n'
+            f'{"Software:":10}{self.software!s}\n'
+            f'{"Context:":10}{(self.context or "DEFAULT")!s}\n'
+        )
+
     def __len__(self):
         return self.local_inputs['length']
 
+    def init_elements(self, elements):
+        self._elements = [Element(self, **i) for i in elements]
+
+    def as_dict(self):
+        'Return attributes dict with preceding underscores removed.'
+        self_dict = {k.lstrip('_'): getattr(self, k) for k in self.__slots__}
+        self_dict.pop('workflow')
+        self_dict['status'] = (self.status.name, self.status.value)
+        self_dict['elements'] = [i.as_dict() for i in self_dict['elements']]
+        for i in [
+            'software_instance',
+            'prepare_software_instance',
+            'process_software_instance',
+        ]:
+            self_dict[i] = self_dict[i].as_dict()
+
+        return self_dict
+
     def generate_id(self):
         self.id = secrets.token_hex(10)
 
+    @property
+    def workflow(self):
+        return self._workflow
+
     @property
     def id(self):
         return self._id
@@ -403,21 +521,76 @@ class Task(object):
     def name(self):
         return self._name
 
+    @property
+    def name_friendly(self):
+        'Capitalise and remove underscores'
+        name = '{}{}'.format(self.name[0].upper(), self.name[1:]).replace('_', ' ')
+        return name
+
     @property
     def method(self):
         return self._method
 
+    @property
+    def elements(self):
+        return tuple(self._elements)
+
     @property
     def software_instance(self):
         return self._software_instance
 
+    @property
+    def prepare_software_instance(self):
+        return self._prepare_software_instance
+
+    @property
+    def process_software_instance(self):
+        return self._process_software_instance
+
     @property
     def task_idx(self):
         return self._task_idx
 
     @property
     def run_options(self):
-        return self._run_options
+        return {**self.software_instance.required_scheduler_options, **self._run_options}
+
+    @property
+    def prepare_run_options(self):
+        return {
+            **self.prepare_software_instance.required_scheduler_options,
+            **self._prepare_run_options,
+        }
+
+    @property
+    def process_run_options(self):
+        return {
+            **self.process_software_instance.required_scheduler_options,
+            **self._process_run_options,
+        }
+
+    def get_scheduler_options(self, task_type='main'):
+        """
+        Parameters
+        ----------
+        task_type : str
+            One of "main", "prepare", "process"
+        """
+        run_opts = {
+            'main': self.run_options,
+            'prepare': self.prepare_run_options,
+            'process': self.process_run_options,
+        }[task_type]
+
+        non_scheduler_opts = ['num_cores', 'job_array']
+        scheduler_opts = {}
+        for k, v in run_opts.items():
+            if k in non_scheduler_opts:
+                continue
+            if k == 'pe':
+                v = v + ' ' + str(run_opts['num_cores'])
+            scheduler_opts.update({k: v})
+        return scheduler_opts
 
     @property
     def status(self):
@@ -425,8 +598,8 @@ class Task(object):
 
     @status.setter
     def status(self, status):
-        'Set task status'
-        # TODO validate, maybe with enum.
+        if status not in TaskStatus:
+            raise TypeError('`status` must be a `TaskStatus`.')
         self._status = status
 
     @property
@@ -441,28 +614,6 @@ class Task(object):
     def local_inputs(self):
         return self._local_inputs
 
-    @property
-    def inputs(self):
-        return self._inputs
-
-    @inputs.setter
-    def inputs(self, inputs):
-        'Set the task inputs (i.e. from `Workflow.prepare_task`).'
-        if not isinstance(inputs, list) or not isinstance(inputs[0], dict):
-            raise ValueError('Inputs must be a list of dict.')
-        self._inputs = inputs
-
-    @property
-    def outputs(self):
-        return self._outputs
-
-    @outputs.setter
-    def outputs(self, outputs):
-        'Set the task outputs (i.e. from `Workflow.process_task`).'
-        if not isinstance(outputs, list) or not isinstance(outputs[0], dict):
-            raise ValueError('Outputs must be a list of dict.')
-        self._outputs = outputs
-
     @property
     def output_map_options(self):
         return self._output_map_options
@@ -471,17 +622,6 @@ class Task(object):
     def schema(self):
         return self._schema
 
-    @property
-    def files(self):
-        return self._files
-
-    @files.setter
-    def files(self, files):
-        'Set the task files (i.e. from `Workflow.prepare_task`).'
-        if not isinstance(files, list) or not isinstance(files[0], dict):
-            raise ValueError('Files must be a list of dict.')
-        self._files = files
-
     @property
     def resource_usage(self):
         return self._resource_usage
@@ -515,11 +655,69 @@ class Task(object):
         return self._merge_priority
 
     @property
-    def name_friendly(self):
-        'Capitalise and remove underscores'
-        name = '{}{}'.format(self.name[0].upper(), self.name[1:]).replace('_', ' ')
-        return name
+    def software(self):
+        return self.software_instance.software
+
+    def get_formatted_commands(self):
+        fmt_commands, input_vars = self.schema.command_group.get_formatted_commands(
+            self.local_inputs['inputs'].keys(),
+            num_cores=self.run_options['num_cores'],
+        )
+
+        # TODO: ?
+        # fmt_commands_new = []
+        # for i in fmt_commands:
+        #     i['line'] = i['line'].replace('<<executable>>', executable)
+        #     fmt_commands_new.append(i)
+        # fmt_commands = fmt_commands_new
+
+        return fmt_commands, input_vars
+
+    def get_prepare_task_commands(self, is_array=False):
+        cmd = f'matflow prepare-task --task-idx={self.task_idx}'
+        cmd += f' --array' if is_array else ''
+        cmds = [cmd]
+        if self.software_instance.task_preparation:
+            env_list = self.software_instance.task_preparation.env.as_list()
+            cmds = env_list + cmds
+        out = [{'subshell': '\n'.join(cmds)}]
+        return out
+
+    def get_prepare_task_element_commands(self, is_array=False):
+        cmd = (f'matflow prepare-task-element --task-idx={self.task_idx} '
+               f'--element-idx=$(($SGE_TASK_ID-1)) '
+               f'--directory={self.workflow.path}')
+        cmd += f' --array' if is_array else ''
+        cmds = [cmd]
+        if self.software_instance.task_preparation:
+            env_list = self.software_instance.task_preparation.env.as_list()
+            cmds = env_list + cmds
+        out = [{'subshell': '\n'.join(cmds)}]
+        return out
+
+    def get_process_task_commands(self, is_array=False):
+        cmd = f'matflow process-task --task-idx={self.task_idx}'
+        cmd += f' --array' if is_array else ''
+        cmds = [cmd]
+        if self.software_instance.task_processing:
+            env_list = self.software_instance.task_processing.env.as_list()
+            cmds = env_list + cmds
+        out = [{'subshell': '\n'.join(cmds)}]
+        return out
+
+    def get_process_task_element_commands(self, is_array=False):
+        cmd = (f'matflow process-task-element --task-idx={self.task_idx} '
+               f'--element-idx=$(($SGE_TASK_ID-1)) '
+               f'--directory={self.workflow.path}')
+        cmd += f' --array' if is_array else ''
+        cmds = [cmd]
+        if self.software_instance.task_processing:
+            env_list = self.software_instance.task_processing.env.as_list()
+            cmds = env_list + cmds
+        out = [{'subshell': '\n'.join(cmds)}]
+        return out
 
     @property
-    def software(self):
-        return self.software_instance['name']
+    def HDF5_path(self):
+        'Get the HDF5 path to this task.'
+        return self.workflow.HDF5_path + f'/\'tasks\'/data_0/data_{self.task_idx}'
